#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <mntent.h>
#include <unistd.h>
#include <string.h>
#include <sys/mount.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/cdrom.h>
#include <linux/fs.h>

#include "cdrom.h"

#include "config.h"

/* misc stuff */
#define True 1
#define False 0

/* mounting defaults */
#ifndef DEVICE
#define DEVICE "/dev/cdrom"
#endif
#ifndef MTAB
#define MTAB "/etc/mtab"
#endif
#ifndef FSTAB
#define FSTAB "/etc/fstab"
#endif

static int mounting_disabled;
static char *cdrom_device;

/* state variables */

static int cdrom_fd;

static int tray_open;
static int cd_in_drive;
static int playing;
static int paused;
static int data_only;
static int contains_data;
static int current_track;
static int mounted;

static time_t last_mtab_change;

typedef struct
{
  struct cdrom_tochdr tochdr;
  struct cdrom_tocentry tocentries[100];
  struct cdrom_subchnl subchnl;
} cdhw_t;

cdhw_t *hw;

int cdrom_read_subchnl()
{
  hw->subchnl.cdsc_format = CDROM_MSF;
  return(ioctl(cdrom_fd, CDROMSUBCHNL, &(hw->subchnl)));
}


int cdrom_update_state()
{
  int i;

  cdrom_check_mounted();
  
  if(mounted) /* If it's mounted, we don't do any ioctl()s, because
		 they'll crash applications accessing the FS. */
    {
#ifdef DEBUG
      fprintf(stderr, "++++ CD is mounted\n");
#endif
      contains_data = True;
      cd_in_drive = True;
      data_only = True; /* We don't know this, but in most cases it
			   will be true */
      tray_open = False;
      playing = False;
      paused = False;
      current_track = 1;

      return 0;
    }
  
  /* Read the CD's table of contents */

  /* The TOC header */

#ifdef NOTOCHDR

  hw->tochdr.cdth_trk0 = 1;
  
  for (i=1; i<=100; i++)
    {
#ifdef DEBUG
      fprintf(stderr, "++++ Reading TOC entry no. %d\n", i);
#endif
      hw->tocentries[i-1].cdte_track = i;
      hw->tocentries[i-1].cdte_format = CDROM_MSF;
      if (ioctl(cdrom_fd, CDROMREADTOCENTRY, &(hw->tocentries[i-1]))
	  != -1)
	hw->tochdr.cdth_trk1 = i;
      else
	{
	  if(i==1) /* Got invalid on the first entry; no CD */
	    {
#ifdef DEBUG
	      fprintf(stderr, "++++ Couldn't read a single TOC entry\n");
	      fprintf(stderr, "++++       Conclusion: No CD in drive\n");
#endif
	      cd_in_drive = False;
	      playing = False;
	      paused = False;
	      data_only = False;
	      contains_data = False;
	      return 0;
	    }
#ifdef DEBUG
	  fprintf(stderr, "++++ Couldn't read TOC entry no. %d\n", i);
	  fprintf(stderr, "++++       Conclusion: CD ends here\n");
#endif
	  break;
	}
    }
#ifdef DEBUG
  fprintf(stderr, "++++ Read TOC entries %d through %d\n",
	  hw->tochdr.cdth_trk0, hw->tochdr.cdth_trk1);
#endif

#else /* !def NOTOCHDR */
  
  if(ioctl(cdrom_fd, CDROMREADTOCHDR, &(hw->tochdr)) == -1)
    {
      /* We can't read the TOC header. This means that there is no CD
	 in the drive. */
#ifdef DEBUG
      fprintf(stderr, "++++ Cannot read TOC header (OK if there's no CD in the drive)\n");
#endif

      cd_in_drive = False;
      playing = False;
      paused = False;
      data_only = False;
      contains_data = False;
      return 0;
    }

  /* The tracks' TOC entries */
  for (i=hw->tochdr.cdth_trk0; i<=hw->tochdr.cdth_trk1; i++) {
#ifdef DEBUG
    fprintf(stderr, "++++ Reading TOC entry no. %d\n", i);
#endif
    hw->tocentries[i-1].cdte_track = i;
    hw->tocentries[i-1].cdte_format = CDROM_MSF;
    if ( ioctl(cdrom_fd, CDROMREADTOCENTRY, &(hw->tocentries[i-1])) == -1 ) {
      /* Huh? We can read the TOC header, but not the single entries?
	 There must be something wrong. */
            fprintf(stderr, "ioctl cdromreadtocentry\n");
            return -1;
    }
  }

#endif /* NOTOCHDR */
  
  /* The leadout track */
  hw->tocentries[hw->tochdr.cdth_trk1].cdte_track = CDROM_LEADOUT;
  hw->tocentries[hw->tochdr.cdth_trk1].cdte_format = CDROM_MSF;
  if (ioctl(cdrom_fd, CDROMREADTOCENTRY, &(hw->tocentries[hw->tochdr.cdth_trk1]
					  )) == -1 ) {
    fprintf(stderr, "ioctl cdromreadtocentry\n");
    return -1;
  }
  
  /* The subchannel */
  if (cdrom_read_subchnl() == -1) {
    fprintf(stderr, "ioctl cdromsubchnl\n");
    return -1;
  }
      
  /* Now let's look at the data: Some drivers let the ioctl's pass OK,
     while there isn't even a CD in the drive. */

  if(hw->tochdr.cdth_trk0 == 0 || hw->tochdr.cdth_trk1 == 0 || 
     hw->tochdr.cdth_trk1 < hw->tochdr.cdth_trk0)
    {
#ifdef DEBUG
      fprintf(stderr, "++++ First track has index %d, "
	              "last track has index %d\n",
	      hw->tochdr.cdth_trk0, hw->tochdr.cdth_trk1);
      fprintf(stderr, "++++ Conclusion: No CD in the drive\n");
#endif
      cd_in_drive = False;
    }
  else
    cd_in_drive = True;
  
  /* Now, what kind of CD is it? */
  if(hw->tochdr.cdth_trk0 == hw->tochdr.cdth_trk1 == 1 &&
     (hw->tocentries[0].cdte_ctrl & CDROM_DATA_TRACK))
    {
#ifdef DEBUG
      fprintf(stderr, "++++ Detected data-only CD-ROM\n");
#endif
      data_only = True;
    }
  else
    {
      data_only = False;
    }

  contains_data = False;
  for(i=hw->tochdr.cdth_trk0; i<=hw->tochdr.cdth_trk1; i++)
    if((hw->tocentries[i-1].cdte_ctrl & CDROM_DATA_TRACK))
      contains_data = True;

#ifdef DEBUG
  if(contains_data && !data_only)
    fprintf(stderr, "++++ Detected a mixed-mode CD-ROM\n");
  else if(!contains_data)
    fprintf(stderr, "++++ Detected an audio-only CD\n");
#endif
  
  /* Is it playing? */

  switch(hw->subchnl.cdsc_audiostatus)
    {
    case CDROM_AUDIO_INVALID:
#ifdef DEBUG
      fprintf(stderr, "++++ Got INVALID audio status\n");
#endif
      /* We can't change any state variables if the CDROM doesn't tell
	 us about them */
      break;
    case CDROM_AUDIO_PLAY:
#ifdef DEBUG
      fprintf(stderr, "++++ CD is playing\n");
#endif
      playing = True;
      paused = False;
      current_track = hw->subchnl.cdsc_trk;
      break;
    case CDROM_AUDIO_PAUSED:
#ifdef DEBUG
      fprintf(stderr, "++++ CD is paused\n");
#endif
      playing = True;
      paused = True;
      current_track = hw->subchnl.cdsc_trk;
      break;
    default:
#ifdef DEBUG
      fprintf(stderr, "++++ CD is neither playing nor paused (status 0x%x)\n", hw->subchnl.cdsc_audiostatus);
#endif
      playing = False;
      paused = False;
      if(!current_track)
	{
	  current_track = hw->tochdr.cdth_trk1;
	  cdrom_skipf_audio();
	}
      
    }

  return 0;
}


int cdrom_tray_open() { return tray_open; }
int cdrom_cd_in_drive() { return cd_in_drive; }
int cdrom_playing() { return playing; }
int cdrom_paused() { return paused; }
int cdrom_cd_is_data_only() { return data_only; }
int cdrom_cd_is_mixed_mode() { return (!data_only && contains_data); }
int cdrom_cd_contains_data() { return contains_data; }

/* This is a bit more complicated: We open /etc/mtab and check for an
   entry containing the configured device. */
int cdrom_mounted() { return mounted; }

void cdrom_check_mounted()
{
  FILE *mtab;
  struct mntent *ent;
  struct stat buf;

  if(stat(MTAB, &buf) < 0)
    {
      fprintf(stderr, "cdrom.c: Couldn't stat() mtab file %s!\n",
	      MTAB);
      perror("cdrom.c: stat");
      fprintf(stderr, "cdrom.c: Giving up.\n");
      exit(1);
    }

  if(last_mtab_change != buf.st_mtime)
    {
      last_mtab_change = buf.st_mtime;
      
      mtab = setmntent(MTAB, "r");
      if(!mtab) /* couldn't access mtab! What's wrong here? */
	{
	  fprintf(stderr, "cdrom.c: Couldn't fopen() mtab file %s!\n",
		  MTAB);
	  perror("cdrom.c: fopen");
	  fprintf(stderr, "cdrom.c: Giving up.\n");
	  exit(1);
	}
      
      while((ent = getmntent(mtab)))
	if(!strcmp(ent->mnt_fsname, cdrom_device))
	  {
	    endmntent(mtab);
	    mounted = 1;
	    return;
	  }

      endmntent(mtab);
      mounted = 0;
    }
}

/* This one returns the number of the current track. Note: The current
   track isn't necessarily playing! (E.g. you can skip
   forward/backward without the CD playing). Returns -1 if no tracks
   can be found (i.e. no CD in the drive).*/

int cdrom_current_track()
{
  if(mounted)
      return -1;
  if(!playing)
    return current_track;
  if(cdrom_read_subchnl() == -1)
    return -1;
  if(hw->subchnl.cdsc_trk < hw->tochdr.cdth_trk0 ||
     hw->subchnl.cdsc_trk > hw->tochdr.cdth_trk1)
    return -1;

  return (current_track = hw->subchnl.cdsc_trk);
}

int cdrom_set_current_track(int track)
{
  if(mounted)
    return -1;
  if(track < hw->tochdr.cdth_trk0 || track > hw->tochdr.cdth_trk1)
    return -1;

  if(playing)
    {
      if(cdrom_track_type(track) == 0) /* Data */
	return -1;

      if(paused)
	{
	  cdrom_stop();
	  current_track = track;
	  cdrom_play();
	  cdrom_pause();
	}
      else
	{
	  cdrom_stop();
	  current_track = track;
	  cdrom_play();
	}
      return 0;
    }
  
  current_track = track;
  return 0;
}

int cdrom_track_type(int track)
{
  if(track < hw->tochdr.cdth_trk0 || track > hw->tochdr.cdth_trk1)
    return -1;

  if((hw->tocentries[track-1].cdte_ctrl & CDROM_DATA_TRACK))
    return 0;
  else
    return 1;
}

int cdrom_get_length_of_track(int track,
			      cdrom_position_t *length_return)
{
  cdrom_position_t endofcd;
  
  if(track < hw->tochdr.cdth_trk0 || track > hw->tochdr.cdth_trk1)
    return -1;

  if(track==hw->tochdr.cdth_trk1)
    {
      if(cdrom_length_of_cd(&endofcd) == -1) return -1;
      length_return->minute = endofcd.minute -
	hw->tocentries[track-1].cdte_addr.msf.minute;
      length_return->second = endofcd.second -
	hw->tocentries[track-1].cdte_addr.msf.second;
      length_return->frame = endofcd.frame -
	hw->tocentries[track-1].cdte_addr.msf.frame;

      if(length_return->frame<0)
	{
	  length_return->frame+=75;
	  length_return->second--;
	}
      if(length_return->second<0)
	{
	  length_return->second+=60;
	  length_return->minute--;
	}
    }
  else /* if(track != hw->tochdr.cdth_trk1) */
    {
      length_return->minute =
	hw->tocentries[track].cdte_addr.msf.minute -
	hw->tocentries[track-1].cdte_addr.msf.minute;
      length_return->second =
	hw->tocentries[track].cdte_addr.msf.second -
	hw->tocentries[track-1].cdte_addr.msf.second;
      length_return->frame =
	hw->tocentries[track].cdte_addr.msf.frame -
	hw->tocentries[track-1].cdte_addr.msf.frame;

      if(length_return->frame<0)
	{
	  length_return->frame+=75;
	  length_return->second--;
	}
      if(length_return->second<0)
	{
	  length_return->second+=60;
	  length_return->minute--;
	}
    }
  return 0;
}

int cdrom_get_position_relative(cdrom_position_t *position_return)
{
  if(mounted) return -1;

  if(!playing)
     {
       position_return->minute = 0;
       position_return->second = 0;
       position_return->frame = 0;
       return 0;
     }

  if(cdrom_read_subchnl() == -1) return -1;

  position_return->minute = hw->subchnl.cdsc_reladdr.msf.minute;
  position_return->second = hw->subchnl.cdsc_reladdr.msf.second;
  position_return->frame = hw->subchnl.cdsc_reladdr.msf.frame;
  return 0;
}

int cdrom_get_position_absolute(cdrom_position_t *position_return)
{
  if(mounted) return -1;

  if(!playing)
     {
       position_return->minute = 0;
       position_return->second = 0;
       position_return->frame = 0;
       return 0;
     }

  if(cdrom_read_subchnl() == -1) return -1;

  position_return->minute = hw->subchnl.cdsc_absaddr.msf.minute;
  position_return->second = hw->subchnl.cdsc_absaddr.msf.second;
  position_return->frame = hw->subchnl.cdsc_absaddr.msf.frame;
  return 0;
}

int cdrom_no_of_tracks()
{
  if(!cd_in_drive)
    return -1;
  return(hw->tochdr.cdth_trk1 - hw->tochdr.cdth_trk0 + 1);
}

int cdrom_length_of_cd(cdrom_position_t *length_return)
{
  if(!cd_in_drive)
    return -1;
  length_return->minute =
    hw->tocentries[hw->tochdr.cdth_trk1].cdte_addr.msf.minute;
  length_return->second =
    hw->tocentries[hw->tochdr.cdth_trk1].cdte_addr.msf.second;
  length_return->frame =
    hw->tocentries[hw->tochdr.cdth_trk1].cdte_addr.msf.frame;
  return 0;
}

/* Actions. All of these return -1 if error, 0 otherwise */

int cdrom_play_absolute(cdrom_position_t pos)
{
  struct cdrom_msf msf;
  cdrom_position_t endofcd;

  if(mounted)
    return -1;
  
  if(!cd_in_drive)
    return -1;

  if(cdrom_length_of_cd(&endofcd) == -1)
    return -1;

  msf.cdmsf_min0 = msf.minute;
  msf.cdmsf_sec0 = msf.second;
  msf.cdmsf_frame0 msf.frame;
  msf.cdmsf_min1 = endofcd.minute;
  msf.cdmsf_sec1 = endofcd.second;
  msf.cdmsf_frame1 = endofcd.frame;

  if(ioctl(cdrom_fd, CDROMPLAYMSF, &msf) == -1)
    return -1;
  playing = True;
  paused = False;
  return 0;
}


int cdrom_play()
{
  struct cdrom_msf msf;
  cdrom_position_t endofcd;

  if(mounted)
    return -1;
  
  if(!cd_in_drive)
    return -1;

  if(cdrom_length_of_cd(&endofcd) == -1)
    return -1;

  msf.cdmsf_min0 = hw->tocentries[current_track-1].cdte_addr.msf.minute;
  msf.cdmsf_sec0 = hw->tocentries[current_track-1].cdte_addr.msf.second;
  msf.cdmsf_frame0 = hw->tocentries[current_track-1].cdte_addr.msf.frame;
  msf.cdmsf_min1 = endofcd.minute;
  msf.cdmsf_sec1 = endofcd.second;
  msf.cdmsf_frame1 = endofcd.frame;

  if(ioctl(cdrom_fd, CDROMPLAYMSF, &msf) == -1)
    return -1;
  playing = True;
  paused = False;
  return 0;
}
  
int cdrom_pause()
{
  if(mounted || !playing)
    return -1;
  if(!paused)
    {
      if(ioctl(cdrom_fd, CDROMPAUSE) == -1)
	return -1;
      paused = True;
      return 0;
    }
  else
    {
      if(ioctl(cdrom_fd, CDROMRESUME) == -1)
	return -1;
      paused = False;
      return 0;
    }
}
	
int cdrom_stop()
{
  if(mounted)
    return -1;
  if(ioctl(cdrom_fd, CDROMSTOP) == -1)
    return -1;
  if(!playing)
    {
      current_track = hw->tochdr.cdth_trk1;
      cdrom_skipf_audio();
    }
  playing = False;
  paused = False;
  return 0;
}

int cdrom_skipf()
{
  int playing_tmp = playing;
  int paused_tmp = paused;
  
  if(mounted || tray_open || !cd_in_drive)
    return -1;
  
  /*  if(playing) cdrom_stop();*/

  current_track++;
  if(current_track < hw->tochdr.cdth_trk0 ||
     current_track > hw->tochdr.cdth_trk1) 
    current_track = hw->tochdr.cdth_trk0;

  if(playing_tmp) cdrom_play();
  if(paused_tmp) cdrom_pause();
  return 0;
}

int cdrom_skipb()
{
  int playing_tmp = playing;
  int paused_tmp = paused;
  
  if(mounted || tray_open || !cd_in_drive)
    return -1;
  
  /*  if(playing) cdrom_stop();*/

  current_track--;
  if(current_track < hw->tochdr.cdth_trk0 ||
     current_track > hw->tochdr.cdth_trk1) 
    current_track = hw->tochdr.cdth_trk1;

  if(playing_tmp) cdrom_play();
  if(paused_tmp) cdrom_pause();
  return 0;
}

int cdrom_skipf_audio()
{
  int playing_tmp = playing;
  int paused_tmp = paused;
  int skipped_once;

  if(mounted || tray_open || !cd_in_drive)
    return -1;
  if(data_only) /* You want me to skip data tracks on a 
		   data only CD? */
    return -1;
  

  skipped_once = False;

  current_track++;
  if(current_track < hw->tochdr.cdth_trk0 ||
     current_track > hw->tochdr.cdth_trk1) 
    {
      current_track = hw->tochdr.cdth_trk0;
      skipped_once = True;
    }

  while(cdrom_track_type(current_track) == 0) /* DATA */
    {
      current_track++;
      if(current_track < hw->tochdr.cdth_trk0 ||
	 current_track > hw->tochdr.cdth_trk1) 
	{
	  if(!skipped_once)
	    {
	      current_track = hw->tochdr.cdth_trk0;
	      skipped_once = True;
	    }
	  else /* Skipped the entire CD once and still couldn't find
		  an audio track */
	    {
	      fprintf(stderr, "Inconsistency in data-only detection\n");
	      data_only = True;
	      return -1;
	    }
	}
    }
  if(playing_tmp) cdrom_play();
  if(paused_tmp) cdrom_pause();
  return 0;
}
    

int cdrom_skipb_audio()
{
  int playing_tmp = playing;
  int paused_tmp = paused;
  int skipped_once;

  if(mounted || tray_open || !cd_in_drive)
    return -1;
  if(data_only) /* You want me to skip data tracks on a 
		   data only CD? */
    return -1;
  
  /*  if(playing) cdrom_stop();*/

  skipped_once = False;

  current_track--;
  if(current_track < hw->tochdr.cdth_trk0 ||
     current_track > hw->tochdr.cdth_trk1) 
    {
      current_track = hw->tochdr.cdth_trk1;
      skipped_once = True;
    }

  while(cdrom_track_type(current_track) == 0) /* DATA */
    {
      current_track--;
      if(current_track < hw->tochdr.cdth_trk0 ||
	 current_track > hw->tochdr.cdth_trk1) 
	{
	  if(!skipped_once)
	    {
	      current_track = hw->tochdr.cdth_trk1;
	      skipped_once = True;
	    }
	  else /* Skipped the entire CD once and still couldn't find
		  an audio track */
	    {
	      fprintf(stderr, "Inconsistency in data-only detection\n");
	      data_only = True;
	      return -1;
	    }
	}
    }
  if(playing_tmp) cdrom_play();
  if(paused_tmp) cdrom_pause();
  return 0;
}
    

int cdrom_mount()
{
  FILE *mtab;
  FILE *fstab;
  struct mntent *ent;
  u_long cdrom_mountflags;
  
  cdrom_check_mounted();
  
#ifdef DEBUG
  fprintf(stderr, "++++ mounting CD-ROM\n");
#endif
  
  if(mounted || mounting_disabled)
    {
#ifdef DEBUG
      if(mounted) fprintf(stderr, "++++ already mounted\n");
      if(mounting_disabled) fprintf(stderr, "++++ mounting disabled\n");
#endif
      return -1;
    }
  if(!(fstab = setmntent(FSTAB, "r")))
    {
      fprintf(stderr, "cdrom.c: fopen(): Couldn't open fstab file %s\n", FSTAB);
      perror("cdrom.c: fopen()");
      fprintf(stderr, "cdrom.c: Giving up.\n");
      exit(1);
    }
  if(!(mtab = setmntent(MTAB, "a+")))
    {
      fprintf(stderr, "cdrom.c: fopen(): Couldn't open mtab file %s\n", MTAB);
      perror("cdrom.c: fopen()");
      fprintf(stderr, "cdrom.c: Giving up.\n");
      endmntent(fstab);
      exit(1);
    }
 
  while((ent = getmntent(fstab)))
    if(!(strcmp(ent->mnt_fsname, cdrom_device)))
      break;

  if(!ent)
    {
      fprintf(stderr, "cdrom.c: Couldn't find an entry for %s in fstab\n", cdrom_device);
      endmntent(mtab); endmntent(fstab);
      return -1;
    }

  cdrom_mountflags = 0xC0ED0000;

  if(strstr(ent->mnt_opts, MNTOPT_RO)) cdrom_mountflags |= MS_RDONLY;
  if(strstr(ent->mnt_opts, MNTOPT_NOSUID))
    cdrom_mountflags |= MS_NOSUID;
  if(strstr(ent->mnt_opts, "noexec")) cdrom_mountflags |= MS_NOEXEC;

#ifdef DEBUG
  fprintf(stderr, "++++ mounting %s on %s (type %s)\n",
	  ent->mnt_fsname, ent->mnt_dir, ent->mnt_type);
#endif
  
  if(mount(ent->mnt_fsname, ent->mnt_dir, ent->mnt_type,
	   cdrom_mountflags, NULL) == -1)
    {
      if(!(cdrom_mountflags & MS_RDONLY)) /* we've tried to mount rw;
					     try again */
	{
	  if(mount(ent->mnt_fsname, ent->mnt_dir, ent->mnt_type,
		   cdrom_mountflags | MS_RDONLY, NULL) == -1)
	    {
	      fprintf(stderr, "cdrom.c: mount(): Couldn't mount %s as specified in %s\n", ent->mnt_fsname, FSTAB);
	      perror("cdrom.c: mount()");
	      fprintf(stderr, "cdrom.c: even read-only failed\n");
	      endmntent(mtab); endmntent(fstab);
	      return -1;
	    }
	}
      else
	{
	  fprintf(stderr, "cdrom.c: mount(): Couldn't mount %s as specified in %s\n", ent->mnt_fsname, FSTAB);
	  perror("cdrom.c: mount()");
	  endmntent(mtab); endmntent(fstab);
	  return -1;
	}      
    }
  
  if(addmntent(mtab, ent))
    {
      fprintf(stderr, "cdrom.c: WARNING: Mounted %s, but couldn't update %s\n", ent->mnt_fsname, FSTAB);
      endmntent(mtab); endmntent(fstab);
      return -1;
    }
  
  endmntent(mtab); endmntent(fstab);
  mounted = True;
  return(0);
}

int cdrom_umount()
{
  FILE *mtab;
  FILE *tmpfile;
  struct mntent *ent;
  char buf[255];

#ifdef DEBUG
  fprintf(stderr, "++++ un-mounting CD-ROM\n");
#endif
  if(!mounted || mounting_disabled)
    {
#ifdef DEBUG
      if(!mounted) fprintf(stderr, "++++ CD-ROM not mounted\n");
      if(mounting_disabled) fprintf(stderr, "++++ mounting disabled\n");
#endif
      return -1;
    }

  if(!(mtab = setmntent(MTAB, "r")))
    {
      fprintf(stderr, "cdrom.c: fopen(): Couldn't open mtab file %s\n", MTAB);
      perror("cdrom.c: fopen()");
      fprintf(stderr, "cdrom.c: Giving up.\n");
      exit(1);
    }

  sprintf(buf, "/etc/wmcd%d", getpid());
  
  if(!(tmpfile = setmntent(buf, "w+")))
    {
      fprintf(stderr, "cdrom.c: fopen(): Couldn't open temp file %s\n", buf);
      perror("cdrom.c: fopen()");
      fprintf(stderr, "cdrom.c: Giving up.\n");
      endmntent(mtab);
      exit(1);
    }

  if(umount(cdrom_device) == -1)
    return -1;

  while((ent = getmntent(mtab)))
    {
      if(strcmp(ent->mnt_fsname, cdrom_device))
	addmntent(tmpfile, ent);
    }

  endmntent(tmpfile);
  endmntent(mtab);
  
  if(rename(buf, MTAB)<0)
    {
      fprintf(stderr, "cdrom.c: rename(): Couldn't move temp file %s to %s\n", buf, MTAB);
      perror("cdrom.c: rename()");
      unlink(buf);
      return -1;
    }

  mounted = False;
  return 0;
}

int cdrom_closetray()
{
  if(mounted)
    return 0;
  if(ioctl(cdrom_fd, CDROMCLOSETRAY) == -1)
    return 0;
  
  tray_open = False;
  current_track = 1; /* bugfix */
  return cdrom_update_state();
}

int cdrom_opentray()
{
  if(mounted)
    return -1;
  if(ioctl(cdrom_fd, CDROMEJECT) == -1)
    return -1;
  
  tray_open = True;
  cd_in_drive = False;
  playing = False;
  paused = False;
  data_only = False;
  contains_data = False;
  return 0;    
}

int init_cdrom(char *device, int enable_mounting)
{
  if(!device)
    cdrom_device = DEVICE;
  else
    cdrom_device = device;

#ifdef MOUNTING
  if(!enable_mounting || !MOUNTING) mounting_disabled = True;
#else
  if(!enable_mounting) mounting_disabled = True;
#endif
  else mounting_disabled = False;

  hw = (cdhw_t *)malloc(sizeof(cdhw_t));

  if((cdrom_fd = open(cdrom_device, O_RDONLY)) < 0)
    {
      fprintf(stderr, "Couldn't open %s\n", cdrom_device);
      perror("cdrom");
      return -1;
    }

  /* Some sane defaults */
  playing = False;
  paused = False;
  data_only = False;
  contains_data = False;
  cd_in_drive = False;
  mounted = False;

  /* First of all, we have to try to close the tray. A lot depends on
     this: If we don't know if the tray is open or closed at starting
     time, we can never decide if it's open or if it's closed and
     there's just no CD in it. */

  if(cdrom_closetray() == -1)
    {
      tray_open = True;
      return 0;
    }

  tray_open = False;
  last_mtab_change = 0;

  current_track = 0;
  
  cdrom_check_mounted();
  return cdrom_update_state();
}







